
#  Copyright (C) 2014, 2020  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'nokogiri'
require 'zip'

module Flow


  ##
  # A Tableau Prep Flow file.
  #
  class Flow < TabClass

    attr_reader :name,         :dir,    :type,      :modtime
    attr_reader :version,      :build,  :platform,  :base
    attr_reader :root
    #--
    attr_reader :datasources,     :datasource 
    attr_reader :datasourceNames, :datasourceUINames, :dataSourceNamesMap
    # attr_reader :orphanDataSources  # i.e. not referenced in any Worksheet
    #--
    # attr_reader :dashboards,      :storyboards,       :worksheets
    # attr_reader :parameters,      :actions
    # attr_reader :valid

    ##
    # Creates a Flow from its file name.
    #
    # == Parameters:
    # flowWithDir
    #   The Flow's file name, the Flow can be a TWB or TWBX file.
    #
    def initialize tflWithDir
      raise ArgumentError.new("ERROR in Flow creation: '#{tflWithDir}' must be a String, is a #{tflWithDir.class} \n ")    unless tflWithDir.is_a? String
      raise ArgumentError.new("ERROR in Flow creation: '#{tflWithDir}' must have an extension of .tfl or .tflx \n ")       unless tflWithDir.upcase.end_with?(".TFL", ".TFLX")
      raise ArgumentError.new("ERROR in Flow creation: '#{tflWithDir}' must must be a file, is a Directory\\Folder \n ")   if File.directory?(tflWithDir)
      raise ArgumentError.new("ERROR in Flow creation: '#{tflWithDir}' cannot be found, must be a Tableau Flow file. \n ") unless File.file?(tflWithDir)
      @valid = false
      if File.file?(tflWithDir) then
        @name    = File.basename(tflWithDir)
        @dir     = File.dirname(File.expand_path(tflWithDir))
        @modtime = File.new(tflWithDir).mtime.strftime("%Y-%m-%d %H:%M:%S")
        case File.extname(tflWithDir).downcase
          when '.tlf'  then processTFL(tflWithDir)
          when '.tflx' then processTFLX(flowWithDir)
        end
      end
    end

    def id
        @id ||= @id = @name.hash
    end

    def build
      @build ||= loadBuild
    end

    def release
      @build ||= loadBuild
    end

    def worksheets
      @worksheets.values
    end

    def worksheet name
      @worksheets[name]
    end

    def worksheetNames
      @worksheets.keys
    end

    def dashboards
      @dashboards.values
    end

    def dashboardNames
      @dashboards.keys
    end

    def dashboard name
      @dashboards[name]
    end

    def actions
      @actions.values
    end

    def actionNames
      @actions.keys
    end

    def datasource name
      @dataSourceNamesMap[name]
    end

    def parameters
      @parameters ||= loadParameters
    end

    def orphanDataSources
      @orphanDataSources ||= identifyOrphandatasoUrceS
    end

    def storyboards
      @storyboards.values
    end

    def storyboardNames
      @storyboards.keys
    end

    def storyboard name
      @storyboards[name]
    end

    private

    def processTFLX(twbxWithDir)
      Zip::File.open(twbxWithDir) do |zip_file|
        twb   = zip_file.glob('*.twb').first
        @ndoc = Nokogiri::XML(twb.get_input_stream)
        @type = :twbx
        processDoc
      end
    end

    def processTFL(twbFile)
      @ndoc = Nokogiri::XML(open(twbFile))
      @type = :twb
      processDoc
    end

    def processDoc
      @valid = true
    end

    def loadBuild
      # - earlier Version, need to confirm when source-build began 
      #   @build   = @ndoc.xpath('/flow/comment()').text.gsub(/^[^0-9]+/,'').strip
      @build = if !@ndoc.at_xpath('/flow/@source-build').nil?
                   @ndoc.at_xpath('/flow/@source-build').text
               else 
                   if @ndoc.at_xpath('/flow/comment()').nil?
                      'not found'
                   else
                      @ndoc.at_xpath('/flow/comment()').text.gsub(/^[^0-9]+/,'').strip
                  end
               end
    end

    def loaddatasources
      # puts "LOAD DATA SOURCES"
      # @dataSourcesNode = @ndoc.at_xpath('//flow/datasources')
      @datasources        = Set.new
      @datasourceNames    = SortedSet.new
      @datasourceUINames  = SortedSet.new
      @dataSourceNamesMap = {}
      datasourceNodes     = @ndoc.xpath('//flow/datasources/datasource')
      # puts "DATASOURCENODES : #{@datasourceNodes.length}"
      datasourceNodes.each do |node|
        datasource = Twb::DataSource.new(node,self)
        @datasources << datasource
        @datasourceNames    << datasource.name
        @datasourceNames    << datasource.uiname
        @datasourceUINames  << datasource.uiname
        @dataSourceNamesMap[datasource.name]   = datasource 
        @dataSourceNamesMap[datasource.uiname] = datasource 
      end
      # puts "DATASOURCES    : #{@datasources.length}"
    end

    def loadWorksheets
      @worksheets  = {}
      hiddenSheets = []
      @ndoc.xpath('//flow/windows/window[@hidden="true"]').each do |hs|
        hiddenSheets << hs['name']
      end
      sheets = @ndoc.xpath('//flow/worksheets/worksheet' ).to_a
      sheets.each do |node|
        sheet = Twb::Worksheet.new(node, self)
        sheet.hidden = hiddenSheets.include? sheet.name 
        @worksheets[sheet.name] = sheet
      end
    end

    def loadDashboards
      @dashesNode = @ndoc.at_xpath('//flow/dashboards')
      @dashboards = {}
      dashes = @ndoc.xpath('//flow/dashboards/dashboard').to_a
      dashes.each do |node|
        unless node.attr('type') == 'storyboard' then
          dashboard = Twb::Dashboard.new(node, @worksheets)
          @dashboards[dashboard.name] = dashboard
        end
      end
    end

    def loadStoryboards
      @storyboards = {}
      boards = @ndoc.xpath("//flow/dashboards/dashboard[@type='storyboard']" ).to_a
      boards.each do |node|
        sheet = Twb::Storyboard.new(node)
        @storyboards[sheet.name] = sheet
      end
    end

    def loadWindows
      @windowsnode = @ndoc.at_xpath("//flow/windows")
      @windows = {}
      windows  = @ndoc.xpath("//flow/windows/window[@name]")
      windows.each do |node|
        window = Twb::Window.new(node)
        @windows[window.name] = window
      end
    end

    def loadActions
      @actions = {}
      actionNodes = @ndoc.xpath("//flow/actions/action")
      actionNodes.each do |anode|
        action = Twb::Action.new(anode, @flownode)
        @actions[action.uiname] = action
      end
    end

    def identifyOrphandatasoUrceS
      sheetDataSources = Set.new
      @worksheets.values.each do |sheet|
        sheet.datasources.each do |ds|
            sheetDataSources << ds.uiname
        end
      end
      @orphanDataSources = @datasourceUINames - sheetDataSources
    end

    def writeTwb(name=@name)
      $f = File.open(name,'w')
      if $f
          $f.puts @ndoc
          $f.close
      end
      return name
    end

    def writeTwbx(name=@name)
        emit  "Writing the Workbook, need implementation"
    end

    # Make sure that the TWB has a <dashboards> node.
    # It's possible for a TWB to have no dashboards, and therefore no <dashboards> node.
    def ensureDashboardsNodeExists
      if @dashesNode.nil?
        @dashesNode = Nokogiri::XML::Node.new "dashboards", @ndoc
        # TODO fix this @dataSourcesNode.add_next_sibling(@dashesNode)
      end
    end

    def ensureWindowsNodeExists
      if @windowsnode.nil?
        @windowsnode = Nokogiri::XML::Node.new "windows", @ndoc
        # TODO fix this @dataSourcesNode.add_next_sibling(@windowsnode)
      end
    end

    def getNewDashboardTitle(t)
      title = t
      if @datasources.include?(title)
        inc = 0
        loop do
          inc+=1
          title = t + ' ' + inc.to_s
          if !@datasources.include?(title)
            break
          end
        end
      end
      return title
    end

    def loadParameters
      @parameters = {}
      paramsDS = ndoc.at_xpath('./flow/datasources/datasource[@name="Parameters"]')
      unless paramsDS.nil?
        paramNodes = paramsDS.xpath('.//column')
        paramNodes.each do |pn|
          parameter = Twb::Parameter.new pn
          @parameters[parameter.name] = parameter
        end
      end
      return @parameters
    end

  end  # class Flow

end  # module Twb