#  calculatedfieldsanalyzer.rb - this Ruby script Copyright 2017 Christopher Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'twb'
require 'csv'

module Twb
module Analysis
module Sheets

  class  WorksheetDataStructureCSVEmitter

    # attr_reader :csvFieldsFileName, :csvRecords
    attr_reader :sheetNames,  :sheetCount 

    @@csvSheetPanesFileType = 'WorksheetPanes'
    @@csvSheetPanesHeader   = ['Record #',
                               'Workbook',           
                               'Workbook Dir',
                               'Worksheet',
                               'Pane #',
                               'Pane #of',
                               'Axis Type',
                               'Axis Field Code',
                               'Axis Field Data Source',
                               'Axis Field Data Source (tech)',
                               'Axis Field Prefix',
                               'Axis Field Name',
                               'Axis Field Suffix'
                              ]

    @@csvSheetFieldsFileType = 'WorksheetFieldsStructure'
    @@csvSheetFieldsHeader   = ['Record #',
                                'Workbook',           
                                'Workbook Dir',
                                'Worksheet',
                                'Pane #',
                                'Data Source',
                                'Data Source (tech)',
                                'Function',
                                'Field Code',
                                'Field Name (tech)',
                                'Field Prefix',
                                'Field Suffix',
                                'Position'
                               ]


    def initialize
      @csvPanesFileName = "Twb#{@@csvSheetPanesFileType}.csv"
      @csvPanesFile     = CSV.open(@csvPanesFileName,'w')
      @csvPanesFile << @@csvSheetPanesHeader
      # --
      @csvFieldsFileName = "Twb#{@@csvSheetFieldsFileType}.csv"
      @csvFieldsFile     = CSV.open(@csvFieldsFileName,'w')
      @csvFieldsFile << @@csvSheetFieldsHeader
      # --
      @twbsCnt   = 0
      @sheetsCnt = 0
      @panesCnt  = 0
      @fieldsCnt = 0
    end

    # def self.csvHeader
    #   @@csvHeader
    # end

    # def self.csvFileType
    #   @@csvFieldsFileType
    # end

    def processTwb twb
      @twb    = case twb
                when String        then Twb::Workbook.new(twb)
                when Twb::Workbook then twb
                else raise ArgumentError.new("ERROR: #{twb} must be a String or Workbook, is a #{twb.class}")
                end
      # --
      @sheetNames = SortedSet.new
      # --
      sheets = @twb.worksheets
      sheets.each do |sheet|
        @sheetNames << sheet.name
        processPanes    sheet
        processRowsCols sheet
      end
    end

    def processPanes sheet
      pnodes   = sheet.node.xpath('.//panes/pane')
      panesCnt = pnodes.length
      pnodes.each do |node|
        id = node.has_attribute?('id') ? node.attribute('id').text : '0'
        #--
        if    node.has_attribute?('x-axis-name')
          axistype  = 'x'
          field     = Twb::CodedField.new node.attribute('x-axis-name').text
        elsif node.has_attribute?('y-axis-name')
          axistype = 'y'
          field     = Twb::CodedField.new node.attribute('y-axis-name').text
        else 
          axistype  = nil
          field     = nil
        end
        #--
        @csvPanesFile << [ @panesCnt+=1,
                           @twb.name,                           # 'Workbook',           
                           @twb.dir,                            # 'Workbook Dir',
                           sheet.name,                          # 'Worksheet',
                           id,                                  # 'Pane #',
                           panesCnt,                            # 'Pane #of',
                           axistype,                            # 'Axis Type',
                           field.nil? ? ' ' : field.code,        # 'Axis Field Code',
                           field.nil? ? ' ' : field.dataSource,  # 'Axis Field Code',
                           field.nil? ? ' ' : field.prefix,      # 'Axis Field Prefix',
                           field.nil? ? ' ' : field.name,        # 'Axis Field Name',
                           field.nil? ? ' ' : field.suffix,      # 'Axis Field Suffix'
                         ]
        processPaneFields sheet.name, node, id
      end
    end

      def processPaneFields sheet, node, paneID
        encodedFields = node.xpath('.//encodings/*')
        # puts "    flds : #{encodedFields.length} "
        fieldPos = 0
        encodedFields.each do |ef|
          function = ef.name
          field    = Twb::CodedField.new(ef.attribute('column').text)
          dsuiname = @twb.datasource(field.dataSource).uiname
          # puts "      ef : <<#{function}>>  #{ef}  -> #{field} "
          @csvFieldsFile << [  @fieldsCnt+=1,     # 'Record #',
                               @twb.name,         # 'Workbook',           
                               @twb.dir,          # 'Workbook Dir',
                               sheet,             # 'Worksheet',
                               paneID,            # 'Pane #',
                               dsuiname,          # 'Data Source',
                               field.dataSource,  # 'Data Source (tech)',
                               function,          # 'Function',
                               field.code,        # 'Field Code',
                               field.name,        # 'Field Name (tech)',
                               field.prefix,      # 'Field Prefix',
                               field.suffix,      # 'Field Suffix',
                               fieldPos+=1
                            ]
        end

      end

      def processRowsCols sheet
        emitRowColFields sheet, 'row', sheet.rowFields
        emitRowColFields sheet, 'col', sheet.colFields
      end

      def emitRowColFields sheet, type, fields
        fieldPos = 0
        fields.each do |field|
          dsuiname = @twb.datasource(field.dataSource).uiname
          @csvFieldsFile << [  @fieldsCnt+=1,     # 'Record #',
                               @twb.name,         # 'Workbook',           
                               @twb.dir,          # 'Workbook Dir',
                               sheet.name,        # 'Worksheet',
                               nil,               # 'Pane #',
                               dsuiname,          # 'Data Source',
                               field.dataSource,  # 'Data Source (tech)',
                               type,              # 'Function',
                               field.code,        # 'Field Code',
                               field.name,        # 'Field Name (tech)',
                               field.prefix,      # 'Field Prefix',
                               field.suffix,      # 'Field Suffix',
                               fieldPos+=1
                            ]
        end
      end

  end # class WorksheetDataStructureCSVEmitter

end   # module Sheets
end   # module Analysis
end   # module Twb