#  sheetfieldsanalyzer.rb Copyright (C) 2018  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.  
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'twb'
require 'csv'

module Twb
module Analysis

  class SheetFieldsAnalyzer

    include TabTool

    @@fieldTypeCodes = { 'DB ref'  => 'Db',
                         'color'   => 'Clr',
                         'filter'  => 'Flt',
                         'lod'     => 'LoD',
                         'page'    => 'Pge',
                         'path'    => 'Pth',
                         'shape'   => 'Shp',
                         'size'    => 'Siz',
                         'text'    => 'Tex',
                         'tooltip' => 'TTp',
                         :column   => 'Clm',
                         :row      => 'Row',
                         'slice'   => 'Slc',
                       }


    attr_accessor :localEmit

    def initialize(**args)
      @args      = args
      @recordDir = !@args.nil? && @args[:recordDir] == true
      @ttdocdir  = @args[:ttdocdir]
      @csvAdd    = args[:csvMode] == :add
      @csvMode   = @csvAdd ? 'a' : 'w'
      # @localEmit = true
      init
      @funcdoc   = {:class=>self.class, :blurb=>'Analyze Worksheet Fields', :description=>nil,}
      #--
      docFileName      = docFile('WorksheetFields.csv')
      @sheetFieldsCSV  = CSV.open(docFileName,@csvMode)
      unless @csvAdd
        if @recordDir
          @sheetFieldsCSV  << ['Rec #','Workbook','Workbook Modified','Worksheet','Data Source','Data Source (tech)','Field','Field (tech)','Usage','Usage - code','Workbook Dir']
        else
          @sheetFieldsCSV  << ['Rec #','Workbook','Workbook Modified','Worksheet','Data Source','Data Source (tech)','Field','Field (tech)','Usage','Usage - code'               ]
        end
      end
      addDocFile @sheetFieldsCSV, docFileName, "Workbooks, Worksheets, and the Sheets' Data Sources and Fields"
      #--
      @twbCnt     = 0
      @sheetCnt   = 0
      @fieldsCnt  = 0
      @recNum     = 0
    end

    def metrics
      {
        # '# of Workbooks'         => @twbcount,
        '# of Worksheets'        => @sheetCnt,
        '# of Worksheet Fields'  => @fieldsCnt
      }
    end

    def processTWB twb
       @twb = twb
       @modTime = @twb.modtime
       emit "   -- twb:: #{@twb.name}"
       @twbCnt += 1
       @twbDomainsLoaded = false
       parseSheets
       finis
    end

    private

    def parseSheets
      @worksheets  = @twb.worksheets
      @worksheets.each do |sheet|
        @sheetName  = sheet.name
        @sheetCnt  += 1
        emit "SHEET: #{@sheetName}"
        showFields sheet unless sheet.datasourceFields.nil?
      end
    end

    def showFields sheet
      @sheetFields = Hash.new { |ds,fields| ds[fields] = Set.new }
      # recordCSV [@twb.name, @sheetName, nil, nil, nil, nil, nil]
      emit "Sheet: #{@sheetName}"
      recordDBFields       sheet
      recordRCFields       sheet.rowFields, :row
      recordRCFields       sheet.colFields, :column
      recordEncodedFields  sheet
      recordFilterFields   sheet
      recordPageFields     sheet
      recordDSFilterFields sheet
      recordSlicesFields   sheet
    end

    def recordDBFields sheet
      fields = sheet.datasourceFields
      emit "def recordDBFields sheet: #{sheet.name} #FIELDS: #{fields.length}"
      fields.each do |dsName, dsfields|
        ds = @twb.datasource dsName
        emit "      ds: #{dsName}"
        emit "        - #{ds.uiname}"
        emit "        : #{ds.class}"
        dsfields.each do |sheetField|
           @fieldsCnt += 1
           emit "       f: #{sheetField}"
           emit "       c: #{sheetField.class}"
           fuiName = ds.fieldUIName sheetField #Fields[sheetField]
           recordCSV [@twb.name, @modTime, @sheetName, ds.uiname, dsName, sheetField.uiname, sheetField.name, 'DB ref', ftc('DB ref')]
        end
      end
    end

    def recordRCFields fields, usage
      emit "def recordRCFields #fields: #{fields.length}  \t #{fields}"
      fields.each do |cf|
        emit "coded field: #{cf}"
        fldName = cf.name
        dsName  = cf.dataSource
        ds      = @twb.datasource dsName
        emit "DATASOURCE : #{ds.class} "  #{ }" #{ds}"
        fuiName = ds.fieldUIName cf.name
        emit "      ds: #{dsName}"
        emit "        - #{ds.uiname}"
        emit "        : #{ds.class}"
        recordCSV [@twb.name, @modTime, @sheetName, ds.uiname, dsName, fuiName, fldName, usage, ftc(usage)]
      end
    end

    def recordEncodedFields sheet
      unless sheet.encodedFields.nil?
        sheet.encodedFields.each do |type,fields|
          fields.each do |field|
            dsName  = field.dataSource
            ds      = @twb.datasource dsName
            fuiName = ds.fieldUIName field.name
            recordCSV [@twb.name, @modTime, @sheetName, ds.uiname, dsName, fuiName, field.name, type, ftc(type)]
          end
        end
      end
    end #def recordEncodedFields

    def recordFilterFields sheet
      filters = sheet.filters
      filters.each do |filter|
        dsName   = filter.dataSource.name
        dsUIName = filter.dataSource.uiname
        recordCSV [@twb.name, @modTime, @sheetName, dsUIName, dsName, filter.uiname, filter.name, 'filter', ftc('filter')]
      end
    end

    def recordPageFields sheet
      sheet.pageFields.each do |pfield|
        ds = @twb.datasource pfield.dataSource
        fuiname = ds.fieldUIName pfield.name
        recordCSV [@twb.name, @modTime, @sheetName, ds.uiname, ds.name, fuiname, pfield.name, 'page', ftc('page')]
      end
    end

    def recordDSFilterFields sheet
      sheet.slicesFields.each do |field|
          dsName   = field.dataSource
          ds       = @twb.datasource dsName
          dsUIName = ds.uiname
          ds.filters.each do |filter|
            recordCSV [@twb.name, @modTime, @sheetName, dsUIName, dsName, filter.uiname, filter.name, 'dsfilter', ftc('dsfilter')]
          end
      end
    end

    def recordSlicesFields sheet
      sheet.slicesFields.each do |field|
        dsName   = field.dataSource
        ds       = @twb.datasource dsName
        dsuiname = ds.uiname
        fuiName  = ds.fieldUIName field.name
        recorded = @sheetFields.key?(dsuiname) && @sheetFields[dsuiname].include?(fuiName)
        emit "      # #{@sheetFields.inspect}"
        emit "      - %-6s  %-25s  -> %-s" % [recorded, dsuiname, fuiName]
        unless recorded
          recordCSV [@twb.name, @modTime, @sheetName, dsuiname, dsName, fuiName, field.name, 'slice', ftc('slice')]
          emit "      - recorded"
        end
        emit "--"
      end
      emit "\n "
    end

    def recordCSV record
      numberedRec = [@recNum+=1] + record
      if @recordDir
        @sheetFieldsCSV  << numberedRec.push(@twb.dir)
      else
        @sheetFieldsCSV  << numberedRec
      end
      @sheetFields[record[2]] << record[4]
    end

    def ftc type   # ftc : abbr for fieldTypeCode, for brevity
      @@fieldTypeCodes[type].nil? ? type : @@fieldTypeCodes[type]
    end

  end #class SheetFieldsAnalyzer

end # module Analysis
end # module Twb
