#  Copyright (C) 2014, 2018  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'nokogiri'
require 'pry'

module Twb

  class ColumnField < TabClass

    include Comparable

    @@dataTypeRoleMap = {
                          'date'     => :dimension,
                          'datetime' => :dimension, 
                          'boolean'  => :dimension,
                          'integer'  => :measure,
                          'float'    => :measure,
                          'real'     => :measure,
                          'string'   => :dimension,
                        }

    # XML Attributes known to exist in TWB xml
    # ---------------------------------------- 
    # aggregation
    # aggregation-param
    # alias
    # auto-column
    # caption
    # datatype
    # datatype-customized
    # default-format
    # default-role
    # default-type
    # hidden
    # layered
    # name
    # param-domain-type
    # pivot
    # role
    # semantic-role
    # SplitFieldIndex
    # SplitFieldOrigin
    # type
    # user-datatype
    # value
    # visual-totals

    attr_reader :node,       :properties
    attr_reader :name,       :caption,       :uiname   
    attr_reader :hasaliases, :alias
    attr_reader :dataType,   :defaultFormat, :paramDomainType
    attr_reader :role,       :semanticRole       
    attr_reader :type,       :value
    attr_reader :alias,      :semanticRole,  :aggregation
    attr_reader :autoColumn, :hidden,        :datatypeCustomized
    attr_reader :calcField
    attr_reader :path
    attr_reader :aliases

    def initialize(fieldNode, datasource=nil)
      @datasource         = datasource
      @node               = fieldNode
      @name               = load('name').sub(/^\[/,'').sub(/\]$/,'')
      @caption            = load 'caption'
      @uiname             = @caption.nil? ? @name : @caption
      #--
      loadAttributes @node
      # @dataType           = load 'datatype'
      # @userDataType       = load 'user-datatype'
      # @defaultFormat      = load 'default-format'
      # @paramDomainType    = load 'param-domain-type'
      # @type               = load 'type'
      # @value              = load 'value'
      # @alias              = load 'alias'
      # @semanticRole       = load 'semantic-role'
      # @aggregation        = load 'aggregation'
      # @autoColumn         = load 'auto-column'
      # @datatypeCustomized = load 'datatype-customized'
    end

    def id
        @id ||= @id = @name.hash
    end    

    def load nodeName
      attr = @node.attribute(nodeName)
      val  = attr.nil? ? nil : attr.text.strip
      return val
    end

    def loadAttributes node
      # puts "augmenting with node:#{node}"
      @node = node
      if node.has_attribute?('aggregation')         then @aggregation        = node['aggregation']         end 
      if node.has_attribute?('alias')               then @alias              = node['alias']               end
      if node.has_attribute?('auto-column')         then @autoColumn         = node['auto-column']         end 
      if node.has_attribute?('datatype')            then @dataType           = node['datatype']            end
      if node.has_attribute?('datatype-customized') then @datatypeCustomized = node['datatype-customized'] end
      if node.has_attribute?('default-format')      then @defaultFormat      = node['default-format']      end
      if node.has_attribute?('param-domain-type')   then @paramDomainType    = node['param-domain-type']   end
      if node.has_attribute?('role')                then @role               = node['role']                end
      if node.has_attribute?('semantic-role')       then @semanticRole       = node['semantic-role']       end
      if node.has_attribute?('type')                then @type               = node['type']                end 
      if node.has_attribute?('user-datatype')       then @userDataType       = node['user-datatype']       end
      if node.has_attribute?('value')               then @value              = node['value']               end
      # loadAttr @aggregation,        'aggregation'
      # loadAttr @alias,              'alias'
      # loadAttr @autoColumn,         'auto-column'
      # loadAttr @dataType,           'datatype'
      # loadAttr @datatypeCustomized, 'datatype-customized'
      # loadAttr @defaultFormat,      'default-format'
      # loadAttr @paramDomainType,    'param-domain-type'
      # loadAttr @role,               'role'
      # loadAttr @semanticRole,       'semantic-role'
      # loadAttr @type,               'type'
      # loadAttr @userDataType,       'user-datatype'
      # loadAttr @value,              'value'
    end

    # def loadAttr property, attribute
    #   # puts " - def loadAttr  attr:#{attribute}:  prop:#{property}:"
    #   if @node.has_attribute? attribute
    #     # puts "   - has"
    #     property = @node[attribute]
    #   # else
    #   #   puts "   - has not"
    #   # end
    # end

    # def role
    #   @role ||= loadRole
    # end

    # def loadRole
    #   role = @node['role']
    #   if role.nil? 
    #     role = @@dataTypeRoleMap[@dataType].nil? ? @dataType : @@dataTypeRoleMap[@dataType]
    #   end
    #   return role
    # end

    def calcField
      @calcField ||= loadCalcField
    end

    def hidden
      @hidden ||= load('hidden').nil? ? false : true
    end

    def loadCalcField
      if !@node.at_xpath('./calculation').nil?
        @calcField = Twb::CalculatedField.new @node, @datasource
      else 
        @calcField = nil
      end
    end

    def path
      @path ||= @node.path.gsub(/\[[0-9]+\]/,'[]')
    end

    def pathbare
      @pathbare ||= path().gsub(/\[\]/,'')
    end

    def comment
      @commentLines ||= loadComment
    end

    def hasaliases
      @hasaliases ||= !@node.at_xpath('./aliases').nil?
    end

    def aliases
      @aliases ||= loadAliases
    end

    def loadAliases
      @aliases = {}
      if hasaliases
        @node.at_xpath('./aliases//alias').each do |an|
          dbvalue = an['key'].to_s.gsub(/^["]+/,'').gsub(/["]+$/,'')
          @aliases[ dbvalue ] = an['value']
        end
      end
      return @aliases
    end

    # COMMENTED FIELDS
    # --------------------------------------
    # <column caption='Calculation1' datatype='real' name='[Calculation_821344020759588865]' role='measure' type='quantitative'>
    #   <calculation class='tableau' formula='[Formula Length] * 1.1' />
    #   <desc>
    #     <formatted-text>
    #       <run>THIS IS A COMMENT</run>
    #     </formatted-text>
    #   </desc>
    # </column>
    # <column datatype='string' name='[Data Source Name (tech) (CalculatedFieldsReferencedFields.csv)]' role='dimension' type='nominal'>
    #   <desc>
    #     <formatted-text>
    #       <run>THIS IS A COMMENT</run>
    #     </formatted-text>
    #   </desc>
    # </column>
    def loadComment
      @commentLines = []
      runs = @node.xpath('./desc/formatted-text/run')
      runs.each do |run|
        lines = run.text.split(/\n/)
        lines.each do |line|
          @commentLines << line
        end
      end
      return @commentLines
    end

    def hasComment?
      loadComment if @commentLines.nil?
      @commentLines.empty?
    end

    def aliases
      @aliases ||= loadAliases
    end

    def loadAliases
      @aliases = {}
      aliasNodes = @node.xpath('.//alias')
      aliasNodes.each do |node|
        key   = node.attribute('key').text.gsub(/^"|"$/,'')
        value = node.attribute('value').text.gsub(/^"|"$/,'')
        @aliases[key] = value
      end
      return @aliases
    end

    def properties
      @properties ||= loadProperties
    end

    def loadProperties
      @properties = {} if @properties.nil?
      @node.attributes.each do |name,attr|
        @properties[name] = attr.value
      end
      @properties[:uiname] = @name
      return @properties
    end

    def to_s
      "n:#{@name}: uin:#{uiname}: "
    end

    def <=>(other)
      @name <=> other.name
    end

    def eql?(other)
      @name == other.name
    end

     def hash
      @name.hash
    end    

  end  # class ColumnField

end
