  #  Copyright (C) 2014, 2015, 2017  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'nokogiri'
require 'digest/md5'
require 'csv'
require 'pry'

module Twb

  class FieldCalculation  < TabClass

    include TabTool

    attr_reader :node,   :field,   :fieldNode,  :dataSource
    attr_reader :name,   :caption, :uiname
    attr_reader :has_formula
    attr_reader :formula         
    attr_reader :formulaUC
    attr_reader :formulaResolved
    attr_reader :formulaFlat 
    attr_reader :formulaFlatResolved
    attr_reader :formulaLines, :formulaResolvedLines, :formulaLinesCount 
    attr_reader :is_tableCalc
    attr_reader :is_lod,   :lodCodePos
    attr_reader :class,    :scopeIsolation
    attr_reader :fields,   :remoteFields, :referencedFields
    attr_reader :comments, :uuid

    # attr_accessor :ttlogfile

    @@tableCalcs = [ 'FIRST',        'INDEX', 'LAST', 'SIZE',
                     'LOOKUP',       'PREVIOUS_VALUE',
                     'RANK',         'RANK_DENSE',    'RANK_MODIFIED', 'RANK_PERCENTILE', 'RANK_UNIQUE',
                     'RUNNING_AVG',  'RUNNING_COUNT', 'RUNNING_MAX',   'RUNNING_MIN',     'RUNNING_SUM',
                     'SCRIPT_BOOL',  'SCRIPT_INT',    'SCRIPT_REAL',   'SCRIPT_STR',
                     'TOTAL',
                     'WINDOW_SUM',   'WINDOW_AVG',    'WINDOW_COUNT',  
                     'WINDOW_MIN',   'WINDOW_MEDIAN', 'WINDOW_MAX',   'WINDOW_PERCENTILE',
                     'WINDOW_STDEV', 'WINDOW_STDEVP',
                     'WINDOW_CORR',  'WINDOW_COVAR',  'WINDOW_COVARP', 
                     'WINDOW_VAR',   'WINDOW_VARP'
                   ]

    def initialize(calcField, datasource=nil)
      raise ArgumentError.new("FieldCalculation must be initialized with a CalculatedField, has been provided with a #{calcField.class}") if calcField.class != Twb::CalculatedField
      init
      # initLogger
      @field        = calcField
      calcNode      = calcField.node
      @istableCalc  = false
      # @localEmit    = false
      emit "FieldCalculation calcNode.nil? :: #{calcNode.nil?} "
      unless calcNode.nil?
        @node         = calcNode.at_xpath('./calculation')
        @fieldNode    = calcField.node
        @dataSource   = datasource
        @class        = attribText(@node, 'class')
        @remoteFields = {}
        if 'categorical-bin'.eql? @class
          @name      = @node.attribute('column').text.gsub(/^\[|\]$/,'')   # assumes the column attribute exists
          @caption   = attribText(@fieldNode, 'caption')
          uiname     = if datasource.nil?
                          @name
                       else
                          datasource.fieldUIName(@name)
                       end
          @uiname       = "#{uiname} <<group>>"
          @has_formula  = true
          @formula      = "grouped <<[#{uiname}]>> values"
          @formulaLines = [ @formula ] 
          @formulaFlat  = @formula 
          @comments     = [ ]
          @is_lod       = false
        else
          @caption   = calcField.caption
          @name      = calcField.name
          @uiname    = calcField.uiname
          #--
          @scopeIsolation = attribText(@node, 'scope-isolation') 
          #-- Formula --
          @has_formula    = @node.has_attribute?('formula')
          if @has_formula
            @formula      = @node.attribute('formula').text.gsub(/\r\n/,"\n")
            emit "\n-- init: #{@formula}"
            @formulaUC    = @formula.upcase
            @formulaLines = formula.split(/\n|\r\n/)
            @formulaFlat  = flattenFormula(@formulaLines)
            @comments     = getComments(@formulaLines)
            @lodCodePos   = @formula =~ /^[ ]*{[ ]*(fixed|include|exclude)[ ]*/i
            @is_lod       = !lodCodePos.nil? && @lodCodePos >= 0
            @is_tableCalc = @@tableCalcs.any? { |tc| @formulaUC.include?(tc) }  #assessTableCalc @formula
            # puts "#{@lodCodePos} \t #{@is_lod} \t #{@is_lod.class} \t => #{@formula}"
          end
        end
      end
    end

    def id
        @id ||= @id = @formulaFlat.hash + calcField.hash
    end

    def uuid
      @uuid ||= @uuid = Digest::MD5.hexdigest(@formula)
    end

    def attribText(node, attribute)
      node.attribute(attribute).nil? ? nil : node.attribute(attribute).text
    end

    def formulaResolved
      @formulaResolved ||= resolveFormula
    end
    
    def resolveFormula
      # emit "\ndef resolveFormula:\n--\n#{@formula}"
      resolved = @formula
      # emit "\t formula:#{resolved}:"
      # parseFormFields # - extracts the fields from the formula; as persisted they're the internal names
      referencedFields.each do |refField|
        # emit "\t refField: "
        resolved.gsub!(refField.techCode,refField.uiCode)
        #   if calcField.techUIdiff
        #    # puts ":::: #{calcField.techCode}  //  #{calcField.uiCode}"
        #     formula = formula.gsub(calcField.techCode,calcField.uiCode)
        #    # puts ":--: #{formula}"
        #   end
      end
      # emit "\t formula:#{resolved}:"
      return resolved
    end

    def referencedFields
      @referencedFields ||= parseFormFields
    end


    def formulaResolvedLines
      formulaResolved.split(/\n|\r\n/)
    end

    def formulaLinesCount
      @formulaLinesCount ||= formulaLines.length
    end

    def enlineResolvedFormula
    end 

    def formulaFlatResolved
      @formulaFlatResolved ||= flattenResolvedFormula
    end

    def flattenResolvedFormula
      formula = formulaResolved
      formula.gsub(/\n/, ' ')
    end

    def flattenFormula lines
      formula = ''
      lines.each do |line|
          # line.strip
          formula += ' ' + line.gsub(/\/\/.*/, '') # unless line =~ /^[ ]*\/\//
      end
      return formula.strip
    end

     
    def getComments lines
      comments = ''
      lines.each do |line|
        if line =~ /\/\// then
          comments += ' ' + line.gsub(/^.*\/\//,'// ')
        end
      end
      return comments.strip
    end

    private


    def pullString chars
      delim1 = chars.shift
      delim2 = delim1+delim1
      field  = delim1
      done   = false
      until done | chars.empty?
        s01 = chars[0..1].join
        if !delim1.eql? chars[0]
          field += chars.shift
        else
          case s01
          when delim2
            field += delim2
            chars.shift(2)
          when delim1
            field += delim1
            chars.shift
            done = true
          else
            field += delim1
            chars.shift
            done = true
          end
        end
      end
    end

    def pullField chars
      # chars = str.split ''
      done  = false
      ds    = ''
      field = ''
      until done
        s01 = chars[0..1].join
        s02 = chars[0..2].join
        if ']'.eql? chars[0]
          case s01
          when ']]'
            field += ']]'
            chars.shift(2)
          when ']'
            field += chars.shift
            done = true
          else
            if '].['.eql?(s02)
              ds = field + ']'
              chars.shift(2)
              # fldstr   = chars.join
              field    = pullField(chars)[:field]
              done     = true
            else
              field += ']'
              chars.shift
              done = true
            end
          end
        else
          field += chars[0]
          chars.shift
        end
      end
      # puts "field: '#{field}' \t\t ds: #{ds}"
      return {:field => field.sub(/\[/,'').sub(/\]$/,''), :ds => ds.sub(/\[/,'').sub(/\]$/,'') }
    end

    def parseFormFields # formula
      @referencedFields = Array.new
      rawFields         = Array.new
      if !@formula.nil? && @formula.include?('[') && @formula.include?(']')
      chars  = formula.split('')
      until chars.empty?
        char0 = chars[0]
        case char0
        when '"', "'"
          pullString(chars)
        when '['
          rawFields << pullField(chars)
        else
          unless chars.nil? | chars.empty?
            chars.shift
          end
        end
      end
      rawFields.each do |rf|
        ds = rf[:ds]
        dataSource = if ''.eql? ds
                       @dataSource
                     else 
                       @dataSource.workbook.datasource(ds)
                     end
        # fieldUIName = dataSource.fieldUIName(rf[:field])
        refField    = ReferencedField.new(rf[:field], dataSource)
        @referencedFields << refField
        end
      end
      return @referencedFields
    end

    def parseFormFieldsx # formula
      rawFields = Set.new
      if !@formula.nil? && @formula.include?('[') && @formula.include?(']')
        noComms   = @formula.gsub(/\/\/.*\r\n/,' ')
        formBase  = noComms.gsub(/\r\n/,' ')
        formLen   = formBase.length
        formChars = formBase.split '' 
        until formChars.empty?
          c = formChars.shift
          case c
            when '"', "'"  
              pullString(formChars, c)
            when '['       
              rawFields << pullField(formChars, ']', @referencedFields)
          end
        end
      end
      @referencedFields = Set.new
      rawFields.each do |rf|
        # @referencedFields << rf
        dataSource = if ''.eql? rf[:ds]
                       @dataSource
                     else 
                       @dataSource.workbook.datasource(rf[:ds])
                     end
        # if dataSource.nil?
        #   binding.pry
        # end
        fieldUIName = dataSource.fieldUIName(rf[:field])
        # binding.pry
        refField = ReferencedField.new(rf[:field], dataSource)
        # binding.pry
        @referencedFields << refField
      end
      return @referencedFields
    end

  end  # class FieldCalculation


  # class CalculationField
  class ReferencedField
    # is a field used in a calculation, resolved into its human-meaningful form

    include Comparable
    include TabTool

    attr_reader :name,       :techCode     
    attr_reader :uiname,     :uiCode 
    attr_reader :dataSource, :dataSourceName, :dataSourceRef, :dataSourceExists
    attr_reader :fqName,     :type
    attr_reader :techUIdiff

    def initialize name, datasource
      # puts "\n\nReferencedField ::  ds: %-25s  |  n: %s  " % [datasource, name]
      @name             = name
      @dataSource       = datasource
      @dataSourceName   = datasource.nil? ? nil : datasource.uiname
      @dataSourceRef    = :local
      @dataSourceExists = true
      @techCode         = "[#{name}]"
      @techUIdiff       = false
      if dataSource.nil?
        # puts 'a'
        @uiname = @name
        @uiCode = @techCode
        @techUIdiff = false
      else # !datasource.nil?
        # puts 'b'
        # puts "b - found uiname for '#{@name}'?: #{!datasource.fieldUIName(@name).nil?} \t is:#{datasource.fieldUIName(@name)} "
        @uiname = datasource.fieldUIName(@name).nil? ? @name : datasource.fieldUIName(@name) 
        @uiCode = @uiname.nil? ? @techCode :  "[#{@uiname}]"
        @techUIdiff = !@techCode.eql?(@uiCode)
        # puts ":b #{datasource.fieldUIName(@name).nil?}  ... #{@name}  ... #{@uiname}"
        # puts "CalculationField ::       uin: %-25s | @name:%-s" % [@uiname,@name]
      end
      @fqName = "#{@dataSourceName}::#{@uiname}"
      @type   = if @dataSource.nil?
                   :CalculatedField
                else
                   @dataSource.calculatedField(@uiname).nil? ? :DatabaseField : :CalculatedField
                end
    end # initialize

    def id
      @id ||= @id = "#{@dataSourceName}::#{@uiname}"
    end

    def <=>(other)
      # myName    = @uiname.nil? ? '' : @uiname
      # otherName = other.uiName.nil? ? "" : other.uiName
      # # puts "#{@uiname} / #{myName} <=>  #{otherName} / #{other.uiName}"
      # # puts "#{@uiname.nil?} // #{other.uiName.nil?}"
      # myName <=> otherName
      @fqName <=> other.fqName
    end

  end # class ReferencedField

end  # module Twb
