#  Copyright (C) 2014, 2018  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'logger'
require 'nokogiri'
require 'digest/md5'
require 'yaml'

module TabTool

  @@licensed = false

  @@TTDOCDIR    = './ttdoc'
  @configFile = './ttconfig.yml'

  attr_accessor :ttdocdir,  :docDir,   :docfiles
  attr_accessor :ttdocdir,  :logger,   :loglevel,  :logfilename
  attr_accessor :uuid,      :type,     :id,        :properties
  attr_reader   :licensed,  :funcdoc,  :metrics,   :alerts 

  @funcdoc     = {:class=>nil, :blurb=>nil, :description=>nil,}
  @docfiles    = []  # should be of form [{:name=>'docFileName',:description=>'doc file description'}]

  # @logfilename = 'TabTools.ttlog'
  @@loglevel    = Logger::DEBUG #INFO
  @localEmit    = false
  # @docDirSet   = false

  def init
    @alerts = []
    initDocDir
    initLogger
  end

  def ttdocdir= docdir
    # puts "def ttdocdir= #{docdir}" 
    @ttdocdir = docdir
    init
  end

  def initDocDir
    # puts "def initDocDir" 
    # return if @@TTDOCDIR.nil?
    # return if ''.eql?($ttdocdir) && ''.eql?(@@TTDOCDIR)
    @docDir = @ttdocdir.nil? ? @@TTDOCDIR : @ttdocdir
    # puts "   @@TTDOCDIR: #{@@TTDOCDIR}"
    # puts "    @ttdocdir: #{@ttdocdir}"
    # puts "    @docDir  : #{@docDir}"
    # puts ""
    return if Dir.exists?(@docDir)
    if File.exists? @docDir  # on the off chance that the doc dir is already present as a file
      @docDir = ''
    else
      Dir.mkdir @docDir
    end
    return @docDir
  end

  def initLogger
    @logger = nil
    if config(:log)
      logFileName   = docFile("#{self.class.to_s.split('::').last}.ttlog")
      @logger       = Logger.new(logFileName)
      @logger.level = Logger::DEBUG
    end
    return @logger
  end

  def docFile name
    @docDir.nil? ? name : "#{@docDir}/#{name}"
  end

  def license= fileName
    @@licensed = fileName.is_a?(String) && File.exist?(fileName)
  end

  def licensed?
    @@licensed
  end

  def funcdoc
    @funcdoc.nil? ? {:class=>'n/a', :blurb=>'generic TabTool blurb', :description=>'A useful Tableau Tool.'} : @funcdoc
  end

  def hasConfig param
    loadConfig if @configParams.nil?
    @configParams[param]  ? true : false
  end

  def config param
    hasConfig(param) ? @configParams[param] : nil
  end

  def loadConfig
    @configParams = if File.exist?('./ttconfig.yml') 
                      YAML.load( File.open('./ttconfig.yml').read )
                    else
                      {}
                    end
  end

  def docfiles
    @docfiles ||= @docfiles = {}
  end

  def addDocFile file, name, description
    @docfiles = [] if @docfiles.nil?
    @docfiles << {:file=>file, :name=>name, :description=>description}
  end

  def closeDocFiles
    emit "Closing doc files"
    unless @docfiles.nil?
      @docfiles.each do |dfh|
        docfile  = dfh[:file]
        fileName = dfh[:name]
        emit "  -- #{dfh}"
        # system     "ls -l '#{fileName}'"
        docfile.close unless docfile.nil?
        # system     "ls -l '#{fileName}'"
        # emit true, "  -- #{dfh[:name]}"
      end
    end
    emit "-----------------"
  end

  def docFileMaxNameLen
    maxlen = 0
    docfiles.each do |f|
      nameLen = f[:name].class == String ? [maxlen, f[:name].length].max : 0
      maxlen = nameLen.nil? ? maxlen : [maxlen,nameLen].max
    end
    return maxlen
  end

  def docfilesdoc(pad=' ',desc='For documentation and generated data see the following:')
    lines   = SortedSet.new
    paddesc = "#{pad}#{desc}"
    unless @docfiles.nil? || @docfiles.empty?
      nameLen = docFileMaxNameLen
      docfiles.each do |dfi|
        lines << "#{pad}- %-#{nameLen}s    %-s " % [ dfi[:name], dfi[:description] ]
      end
    end
    docLines = lines.empty? || paddesc =~ /^[ ]+/ ? [] : [' ',"#{pad}#{desc}"]
    lines.each do |l|
      docLines << l
    end
    return docLines
  end

  def docfilesdocto_s(pad=' ',desc='For documentation and generated data see the following:')
    str = ''
    docfilesdoc(pad,desc).each { |l| str += "#{l}\n"}
    return str
  end

  # def metrics
  #   {}
  # end

  def initCSV(fileName, desc=nil, header=nil)
    emit "def initCSV"
    csvName =  docFile(fileName)
    emit "     @recordDir: #{@recordDir.nil?} \t :: #{@recordDir}"
    emit "         header: #{@header.nil?} \t :: #{header}"
    emit "        csvName: #{csvName.nil?} :: #{csvName}"
    csvFile = CSV.open(csvName, 'w')
    unless header.nil?
      if @recordDir
        csvFile << header.push('Workbook Directory')
      else
        csvFile << header
      end
    end
    addDocFile csvFile, csvName, desc
    return csvFile
  end

  def emit(local=@localEmit, stuff)
      # if @logger.closed? 
      #   @logger.reopen 
      # end
      if stuff.is_a? String then
        lines = stuff.split(/\n/)
        lines.each do |line|
          @logger.debug "#{@emitPrefix}#{line}" unless @logger.nil?
                   puts "#{@emitPrefix}#{line}" if local
        end
      else
          @logger.debug "#{@emitPrefix}#{stuff}" unless @logger.nil?
                   puts "#{@emitPrefix}#{stuff}" if local
      end
  end

  def emitCSV csvFile, record
    emit record
    csvFile << record unless csvFile.nil?
    emit "##### ALERT ##### CSV file expected, is nil" if csvFile.nil?
  end

  def alert str
    @alerts << str
    emit "#{self.class} :: #{str}"
  end

  def finis
    # @logger.close unless @logger.nil? || @logger.closed?
  end

end # module TabTool

