#  Copyright (C) 2014, 2018  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'nokogiri'
require 'zip'

module Twb


  ##
  # A Tableau Workbook and its parts.
  #
  class Workbook < TabClass

    attr_reader :workbooknode, :node,   :ndoc
    attr_reader :name,         :dir,    :type,      :modtime
    attr_reader :version,      :build,  :platform,  :base
    #--
    attr_reader :datasources,     :datasource 
    attr_reader :datasourceNames, :datasourceUINames, :dataSourceNamesMap
    attr_reader :orphanDataSources  # i.e. not referenced in any Worksheet
    #--
    attr_reader :dashboards,      :storyboards,       :worksheets
    attr_reader :parameters,      :actions
    attr_reader :valid

    ##
    # Creates a Workbook from its file name.
    #
    # == Parameters:
    # twbWithDir
    #   The Workbook's file name, the Workbook can be a TWB or TWBX file.
    #
    def initialize twbWithDir
      raise ArgumentError.new("ERROR in Workbok creation: '#{twbWithDir}' must be a String, is a #{twbWithDir.class} \n ")        unless twbWithDir.is_a? String
      raise ArgumentError.new("ERROR in Workbok creation: '#{twbWithDir}' must have an extension of .twb or .twbx \n ")           unless twbWithDir.upcase.end_with?(".TWB", ".TWBX")
      raise ArgumentError.new("ERROR in Workbok creation: '#{twbWithDir}' must must be a file, is a Directory\\Folder \n ")           if File.directory?(twbWithDir)
      raise ArgumentError.new("ERROR in Workbok creation: '#{twbWithDir}' cannot be found, must be a Tableau Workbook file. \n ") unless File.file?(twbWithDir)
      @valid = false
      if File.file?(twbWithDir) then
        @name    = File.basename(twbWithDir)
        @dir     = File.dirname(File.expand_path(twbWithDir))
        @modtime = File.new(twbWithDir).mtime.strftime("%Y-%m-%d %H:%M:%S")
        case File.extname(twbWithDir)
          when '.twb'  then processTWB(twbWithDir)
          when '.twbx' then processTWBX(twbWithDir)
        end
      end
    end

    def id
        @id ||= @id = @name.hash
    end

    def build
      @build ||= loadBuild
    end

    def release
      @build ||= loadBuild
    end

    def worksheets
      @worksheets.values
    end

    def worksheet name
      @worksheets[name]
    end

    def worksheetNames
      @worksheets.keys
    end

    def dashboards
      @dashboards.values
    end

    def dashboardNames
      @dashboards.keys
    end

    def dashboard name
      @dashboards[name]
    end

    def actions
      @actions.values
    end

    def actionNames
      @actions.keys
    end

    def datasource name
      @dataSourceNamesMap[name]
    end

    def parameters
      @parameters ||= loadParameters
    end

    def orphanDataSources
      @orphanDataSources ||= identifyOrphandatasoUrceS
    end

    def storyboards
      @storyboards.values
    end

    def storyboardNames
      @storyboards.keys
    end

    def storyboard name
      @storyboards[name]
    end

    # Add a new Documentation Dashboard to the TWB.
    # Ensure that the TWB has a <dashboards> node (it may not).
    # Make sure that the new Doc Dashboard's name doesn't conflict with an existing Dashboard - increment the incoming name if necessary.
    # Add Doc Dashboard's <dashboard> and <window> nodes to the TWB; there's always a <windows> node in the TWB.
    def addDocDashboard docDashboard
      ensureDashboardsNodeExists
      ensureWindowsNodeExists
      title = getNewDashboardTitle(docDashboard.title)
      docDashboard.title=(title) unless title == docDashboard.title
      @dashesNode.add_child(docDashboard.dashnode)
      @windowsnode.add_child(docDashboard.winnode)
    end

    # Write the TWB to a file, with an optional name.
    # Can be used to write over the existing TWB (dangerous), or to a new file (preferred).
    def write(name=@name)
      case @type
      when :twb 
        writeTwb(name)
      when :twbx 
        writeTwbx(name)
      else
        emit  "Cannot write this Workbook - it has an invalid type: #{@type}"
        raise "Cannot write this Workbook - it has an invalid type: #{@type}"
      end
    end

    # Write the TWB to a file, appending the base name with the provided string.
    # Intended for use when making adjustments to the TWB without overwriting the original.
    def writeAppend(str)
      newName = @name.sub(/[.]twb$/,'') + str.gsub(/^[.]*/,'.') + '.twb'
      write newName
    end

    private

    def processTWBX(twbxWithDir)
      Zip::File.open(twbxWithDir) do |zip_file|
        twb   = zip_file.glob('*.twb').first
        @ndoc = Nokogiri::XML(twb.get_input_stream)
        @type = :twbx
        processDoc
      end
    end

    def processTWB(twbFile)
      @ndoc = Nokogiri::XML(open(twbFile))
      @type = :twb
      processDoc
    end

    def processDoc
      @node         = @ndoc.at_xpath('//workbook')
      @workbooknode = @node
      @version      = @node.nil? ? nil : @node["version"]
      @platform     = @node.nil? ? nil : @node["source-platform"].nil? ? nil : @node["source-platform"].capitalize
      @base         = @node.nil? ? nil : @node["base"]
      loaddatasources
      loadWorksheets
      loadDashboards
      loadStoryboards
      loadWindows
      loadActions
      @valid = true
    end

    def loadBuild
      # - earlier Version, need to confirm when source-build began 
      #   @build   = @ndoc.xpath('/workbook/comment()').text.gsub(/^[^0-9]+/,'').strip
      @build = if !@ndoc.at_xpath('/workbook/@source-build').nil?
                   @ndoc.at_xpath('/workbook/@source-build').text
               else 
                   if @ndoc.at_xpath('/workbook/comment()').nil?
                      'not found'
                   else
                      @ndoc.at_xpath('/workbook/comment()').text.gsub(/^[^0-9]+/,'').strip
                  end
               end
    end

    def loaddatasources
      # puts "LOAD DATA SOURCES"
      # @dataSourcesNode = @ndoc.at_xpath('//workbook/datasources')
      @datasources        = Set.new
      @datasourceNames    = SortedSet.new
      @datasourceUINames  = SortedSet.new
      @dataSourceNamesMap = {}
      datasourceNodes     = @ndoc.xpath('//workbook/datasources/datasource')
      # puts "DATASOURCENODES : #{@datasourceNodes.length}"
      datasourceNodes.each do |node|
        datasource = Twb::DataSource.new(node,self)
        @datasources << datasource
        @datasourceNames    << datasource.name
        @datasourceNames    << datasource.uiname
        @datasourceUINames  << datasource.uiname
        @dataSourceNamesMap[datasource.name]   = datasource 
        @dataSourceNamesMap[datasource.uiname] = datasource 
      end
      # puts "DATASOURCES    : #{@datasources.length}"
    end

    def loadWorksheets
      @worksheets  = {}
      hiddenSheets = []
      @ndoc.xpath('//workbook/windows/window[@hidden="true"]').each do |hs|
        hiddenSheets << hs['name']
      end
      sheets = @ndoc.xpath('//workbook/worksheets/worksheet' ).to_a
      sheets.each do |node|
        sheet = Twb::Worksheet.new(node, self)
        sheet.hidden = hiddenSheets.include? sheet.name 
        @worksheets[sheet.name] = sheet
      end
    end

    def loadDashboards
      @dashesNode = @ndoc.at_xpath('//workbook/dashboards')
      @dashboards = {}
      dashes = @ndoc.xpath('//workbook/dashboards/dashboard').to_a
      dashes.each do |node|
        unless node.attr('type') == 'storyboard' then
          dashboard = Twb::Dashboard.new(node, @worksheets)
          @dashboards[dashboard.name] = dashboard
        end
      end
    end

    def loadStoryboards
      @storyboards = {}
      boards = @ndoc.xpath("//workbook/dashboards/dashboard[@type='storyboard']" ).to_a
      boards.each do |node|
        sheet = Twb::Storyboard.new(node)
        @storyboards[sheet.name] = sheet
      end
    end

    def loadWindows
      @windowsnode = @ndoc.at_xpath("//workbook/windows")
      @windows = {}
      windows  = @ndoc.xpath("//workbook/windows/window[@name]")
      windows.each do |node|
        window = Twb::Window.new(node)
        @windows[window.name] = window
      end
    end

    def loadActions
      @actions = {}
      actionNodes = @ndoc.xpath("//workbook/actions/action")
      actionNodes.each do |anode|
        action = Twb::Action.new(anode, @workbooknode)
        @actions[action.uiname] = action
      end
    end

    def identifyOrphandatasoUrceS
      sheetDataSources = Set.new
      @worksheets.values.each do |sheet|
        sheet.datasources.each do |ds|
            sheetDataSources << ds.uiname
        end
      end
      @orphanDataSources = @datasourceUINames - sheetDataSources
    end

    def writeTwb(name=@name)
      $f = File.open(name,'w')
      if $f
          $f.puts @ndoc
          $f.close
      end
      return name
    end

    def writeTwbx(name=@name)
        emit  "Writing the Workbook, need implementation"
    end

    # Make sure that the TWB has a <dashboards> node.
    # It's possible for a TWB to have no dashboards, and therefore no <dashboards> node.
    def ensureDashboardsNodeExists
      if @dashesNode.nil?
        @dashesNode = Nokogiri::XML::Node.new "dashboards", @ndoc
        # TODO fix this @dataSourcesNode.add_next_sibling(@dashesNode)
      end
    end

    def ensureWindowsNodeExists
      if @windowsnode.nil?
        @windowsnode = Nokogiri::XML::Node.new "windows", @ndoc
        # TODO fix this @dataSourcesNode.add_next_sibling(@windowsnode)
      end
    end

    def getNewDashboardTitle(t)
      title = t
      if @datasources.include?(title)
        inc = 0
        loop do
          inc+=1
          title = t + ' ' + inc.to_s
          if !@datasources.include?(title)
            break
          end
        end
      end
      return title
    end

    def loadParameters
      @parameters = {}
      paramsDS = ndoc.at_xpath('./workbook/datasources/datasource[@name="Parameters"]')
      unless paramsDS.nil?
        paramNodes = paramsDS.xpath('.//column')
        paramNodes.each do |pn|
          parameter = Twb::Parameter.new pn
          @parameters[parameter.name] = parameter
        end
      end
      return @parameters
    end

  end  # class Workbook

end  # module Twb