#  Copyright (C) 2014, 2018  Chris Gerrard
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require 'nokogiri'
require 'digest/md5'

module Twb

  class Worksheet

    include TabTool

    @@hasher = Digest::SHA256.new

    @fieldEncodingsXPath = './table/panes/pane//encodings'

    attr_reader   :node,   :name,      :datasourcenames, :datasources
    attr_reader   :panesCount
    attr_reader   :fields, :rowFields, :colFields, :paneFields, :datasourceFields, :pageFields, :encodedFields, :slicesFields
    attr_reader   :filters
    attr_reader   :tooltip
    attr_accessor :hidden, :visible

    def initialize sheetNode, twb
      @twb  = twb
      @node = sheetNode
      @name = @node['name']
      emit "########################## Worksheet initialize name: #{@name}"
      loadDataSources
      loadFields
      return self
    end

    def id
        @id ||= @name.hash
    end

    def workbook
      @workbook ||= @node.at_xpath('/').root
    end

    def datasources
      @datasources.values
    end

    def datasource name
      @datasources[name]
    end

    def datasourceFields
      @datasourceFields ||= loadFields
    end

    def fields
      @fields ||= loadFields
      # @fields.nil? ? loadFields : @fields
    end

    def slicesFields
      @slicesFields ||= loadSlicesFields
    end

    def filters
      @filters ||= loadFilters
    end

    def tooltip
      @tooltip ||= loadTooltip
    end

    # def hidden
    #   @hidden ||= resolveHidden
    # end

    def visible
      @visible ||= !hidden
    end

    def addDSFields fields, usage
      fields.each do  
      end
    end

    def panesCount
      @panesCount ||= @node.xpath('.//panes/pane').length
    end

    def paneFields
      @paneFields ||= loadPaneFields
    end

    def pageFields
      @pageFields ||= loadpageFields
    end

    def datasourcenames
      @datasources.keys
    end

    private

    # def resolveHidden
    #   xpath = "//windows/window[@name=\"#{@name}\"]"
    #   emit true, "resolveHidden: #{xpath}"
    #   windowNode = @node.at_xpath(xpath)
    #   @hidden    = !windowNode.nil? && 'true' == windowNode['hidden']
    # end

    def loadFields
      # puts "WORKSHEET loadFields"
      @fields = {}
      @datasourceFields = {}
      dsNodes = @node.xpath('.//datasource-dependencies')
      dsNodes.each do |dsn|
        dsName     = dsn.attribute('datasource').text
        dsFields   = @datasourceFields[dsName] = Set.new
        fieldNodes = dsn.xpath('./column')
        fieldNodes.each do |fn|
          dsFields.add WorksheetField.new fn, dsName
        end
      end
      @fields[:dsfields] = @datasourceFields
      # <rows>([DATA Connection (copy 2)].[none:JIRA HARVEST Correspondence  Name:nk] / [DATA Connection (copy 2)].[none:CreatedDate (Issue) (DAY):ok])</rows>
      # <cols>[DATA Connection (copy 2)].[:Measure Names]</cols>
      @rowFields     ||= loadRowColFields(:rows) # returns map of data source => (Set of) field names
      @fields[:rows]   = @rowFields
      @colFields     ||= loadRowColFields(:cols) # returns map of data source => (Set of) field names
      @fields[:cols]   = @colFields
      loadFieldEncodings
    end

    def loadPaneFields
      @paneFields = Set.new
      panes = @node.xpath('.//pane')
    end

    def loadFilters
      emit ""########################## loadFilters"
      @filters    = []
      filterNodes = @node.xpath('./table/view//filter[@column]')
      filterNodes.each do |fnode|
        filter = QuickFilter.new(fnode,@twb)
        @filters << filter
      end
      return @filters
    end

    def loadDataSources
      @datasources = {}
      dsNodes = @node.xpath('.//datasource')
      dsNodes.each do |dsn|
        ds = WorksheetDataSource.new dsn
        @datasources[ds.name] = ds
      end
    end

    def loadRowColFields(type)
      fields = []
      xpath = case type
        when :rows then './table/rows'
        when :cols then './table/cols'
      end
      return fields if xpath.nil?
      node = @node.at_xpath(xpath)
      emit "def loadRowColFields:: type: #{type} \t node: #{node.class} \t  path: #{node.nil? ? '<<nil>>' : node.path }"
      emit "node      : #{node.inspect}"
      nodeFields = RowsColsSplitter.new(node).fields
      emit "nodeFields: #{nodeFields}"
      nodeFields.each do |fcode|
        fields << CodedField.new(fcode)
      end
      return fields
    end

    def loadTooltip
      @tooltip = @node.xpath('./table/panes/pane/customized-tooltip')
    end

    def loadpageFields 
      @pageFields = [] # { |h,k| h[k] = [] }
      nodes = @node.xpath('.//table/pages/column')
      nodes.each do |node|
        @pageFields << CodedField.new(node.text)
      end
      return @pageFields
    end

    def loadSlicesFields
      @slicesFields = []
      nodes = @node.xpath('./table/view/slices//column')
      nodes.each do |node|
        # puts "Slice field: #{node.text}"
        @slicesFields << CodedField.new(node.text)
      end
      return @slicesFields
    end

    def loadFieldEncodings
      @encodedFields = Hash.new { |h,k| h[k] = [] }
      enodes = @node.xpath('.//table/panes/pane/encodings')
      enodes.each do |enode|
        enode.children.each do |child|
          unless child['column'].nil?
            @encodedFields[child.name] << CodedField.new(child['column'])
          end
        end
      end
      @encodedFields.each do |type, fields|
        @fields[type] = fields
      end
    end

    # def loadFieldEncodings node
    #   $encodedFields = Hash.new { |h,k| h[k] = [] }
    #   enodes = node.xpath('.//table/panes/pane/encodings')
    #   enodes.each do |enode|
    #     enode.children.each do |child|
    #       unless child['column'].nil?
    #         $encodedFields[child.name] << CodedField.new(child['column'])
    #       end
    #     end
    #   end
    #   $encodedFields.each do |type, fields|
    #     $fields[type] = fields
    #   end
    # end


  
  end # class Worksheet

  class WorksheetDataSource
    # --
    attr_reader :node, :name, :caption, :uiname
    # --
    def initialize node
      @node   = node
      @caption = node.attr('caption')
      @name    = node.attr('name').gsub(/^\[/,'').gsub(/\]$/,'')
      @uiname  = if @caption.nil? || @caption == '' then @name else @caption end
    end
  end # class WorksheetDataSource


  class WorksheetField
    # --
    attr_reader :datasource, :name, :caption, :uiname, :node
    # --
    def initialize node, dsName
      @node    = node
      @name    = node.attr('name').gsub(/^\[/,'').gsub(/\]$/,'')
      @caption = node.attr('caption').nil? ? nil : node.attr('caption').gsub(/^\[/,'').gsub(/\]$/,'') 
      @uiname  = if @caption.nil? || @caption == '' then @name else @caption end
      @datasource = dsName
    end
    # --
    def to_s
      "name:#{@name}|caption:#{@caption}|uiname:#{@uiname}||ds:#{datasource}"
    end
  end # WorksheetField


  class RowsColsSplitter
    include Comparable
    attr_reader :fields
    def initialize node
      @fields = []
      return fields if node.nil?
      unless node.text.nil?
        codes = node.text.split(/[\])] [\/+*] [(]?\[/)
        codes.each do |c| 
          @fields << c.gsub(/^[(]*[\[]*|[)\]]*$/,'')
        end
      end
    end
  end # class RowsColsSplitter


end # module Twb